use anyhow::Result;
use std::collections::{HashMap, HashSet};

static INPUT: &str = include_str!("../../input07");

fn main() -> Result<()> {
    let rules = BagRules::from(INPUT);

    let part1 = rules
        .resolve_can_contain("shiny gold")
        .unwrap()
        .iter()
        .count();
    println!("there are {} kinds of bags to look at.", part1);

    let part2 = rules.count_contained_bags("shiny gold");
    println!("there are {} bags in a shiny gold bag.", part2);
    Ok(())
}

type Bag<'a> = &'a str;

#[derive(Debug)]
struct BagRules<'a> {
    inner: HashMap<Bag<'a>, Vec<Bag<'a>>>,
    reverse: HashMap<Bag<'a>, HashSet<Bag<'a>>>,
}

impl<'a> From<&'a str> for BagRules<'a> {
    fn from(data: &'a str) -> Self {
        let inner: HashMap<Bag<'a>, Vec<Bag<'a>>> =
            data.lines().map(BagRules::parse_bag_rule_line).collect();
        let mut reverse = HashMap::default();

        // building the reverse is rather expensive
        inner.iter().for_each(|(key, values)| {
            values.iter().for_each(|&value| {
                let set: &mut HashSet<&'a str> = reverse.entry(value).or_default();
                set.insert(key);
            })
        });
        BagRules { inner, reverse }
    }
}

impl<'a> BagRules<'a> {
    /// returns a list of all the bags that can contain a `bag`
    fn resolve_can_contain(&self, bag: &str) -> Option<HashSet<&'a str>> {
        let mut results: HashSet<&'a str> = HashSet::default();
        let bags = self.reverse.get(bag)?;
        for bag in bags.iter() {
            results.insert(bag);
            if let Some(bags) = self.resolve_can_contain(bag) {
                results.extend(bags);
            };
        }
        Some(results)
    }

    fn count_contained_bags(&self, bag: &str) -> usize {
        // we need to remove the shiny gold back itself from the count
        self.inner_count_contained_bags(bag) - 1
    }

    fn inner_count_contained_bags(&self, bag: &str) -> usize {
        let childs = match self.inner.get(bag) {
            Some(childs) => childs,
            None => panic!("invalid bag found"),
        };
        let bags: usize = childs
            .iter()
            .map(|bag| self.inner_count_contained_bags(bag))
            .sum();
        bags + 1
    }

    fn parse_bag_rule_line(line: &'a str) -> (Bag<'a>, Vec<Bag<'a>>) {
        let mut splitted = line.splitn(2, " bags contain ");
        let key = splitted.next().expect("invalid bag rule line: missing key");
        let values = splitted.next().expect("invalid bag rule line: missing value");

        if values == "no other bags." {
            return (key, vec![]);
        }

        let values: Vec<&str> = values
            .strip_suffix(".")
            .expect("invalid bag rule line: not terminated with `.`")
            .split(',')
            .flat_map(BagRules::parse_bag_content)
            .collect();
        (key, values)
    }

    fn parse_bag_content(value: &'a str) -> Vec<Bag<'a>> {
        let value = value.trim();
        let split_at = value.trim().chars().take_while(|c| c.is_numeric()).count();
        let (count, bag) = value.split_at(split_at);
        let bag = bag
            .strip_suffix(" bags")
            .or(bag.strip_suffix(" bag"))
            .expect("invalid bag content: no bag(s) suffix");
        let count = count.trim().parse::<usize>().unwrap();
        vec![bag.trim(); count]
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn examples2_old_data() {
        let input = "
        light red bags contain 1 bright white bag, 2 muted yellow bags.
dark orange bags contain 3 bright white bags, 4 muted yellow bags.
bright white bags contain 1 shiny gold bag.
muted yellow bags contain 2 shiny gold bags, 9 faded blue bags.
shiny gold bags contain 1 dark olive bag, 2 vibrant plum bags.
dark olive bags contain 3 faded blue bags, 4 dotted black bags.
vibrant plum bags contain 5 faded blue bags, 6 dotted black bags.
faded blue bags contain no other bags.
dotted black bags contain no other bags.
"
        .trim();
        let rules = BagRules::from(input);
        assert_eq!(rules.resolve_contains("shiny gold"), 32)
    }

    #[test]
    fn examples2_new_adata() {
        let input = "
        shiny gold bags contain 2 dark red bags.
dark red bags contain 2 dark orange bags.
dark orange bags contain 2 dark yellow bags.
dark yellow bags contain 2 dark green bags.
dark green bags contain 2 dark blue bags.
dark blue bags contain 2 dark violet bags.
dark violet bags contain no other bags.
"
        .trim();
        let rules = BagRules::from(input);
        assert_eq!(rules.resolve_contains("shiny gold"), 126)
    }
}
